#include <stdio.h>
#include <string.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/muimaster.h>
#include <libraries/asl.h>
#include <libraries/mui.h>
#include <mui/texteditor_mcc.h>

#include "General.h"
#include "ReqMenuitem-Demo.h"

VOID __saveds LoadCode (REG(a0) struct Hook *hook, REG(a2) Object *obj, REG(a1) BPTR *fh)
{
	struct FileInfoBlock *finfo;
	if(finfo = (struct FileInfoBlock *)AllocDosObject(DOS_FIB, NULL))
	{
		if(ExamineFH(*fh, finfo))
		{
			STRPTR buf;
			if(buf = new UBYTE [finfo->fib_Size + 1])
			{
				buf[Read(*fh, buf, finfo->fib_Size)] = '\0';
				SetAttrs(obj,
					MUIA_TextEditor_Contents, buf,
					MUIA_TextEditor_HasChanged, FALSE,
					TAG_DONE);
				delete buf;
			}
		}
		FreeDosObject(DOS_FIB, finfo);
	}
}

VOID __saveds SaveCode (REG(a0) struct Hook *hook, REG(a2) Object *obj, REG(a1) BPTR *fh)
{
	STRPTR buf;
	if(buf = (STRPTR)DoMethod(obj, MUIM_TextEditor_ExportText))
	{
		Write(*fh, buf, strlen(buf));
		set(obj, MUIA_TextEditor_HasChanged, FALSE);
		FreeVec(buf);
	}
}

struct Hook LoadHook = { { NULL, NULL }, (ULONG(*)())LoadCode };
struct Hook SaveHook = { { NULL, NULL }, (ULONG(*)())SaveCode };

VOID main ()
{
	Object *app, *win, *mi_Open, *mi_Save, *mi_SaveAs, *mi_Quit;
	Object *text, *texteditor, *slider;

	app = ApplicationObject,

		MUIA_Application_Author,		"Allan Odgaard",
		MUIA_Application_Base,			PROGRAM,
		MUIA_Application_Copyright,	"1999 Allan Odgaard",
		MUIA_Application_Description,	PROGRAM" - Just a test",
		MUIA_Application_Title,			PROGRAM,
		MUIA_Application_Version,		"$VER: " PROGRAM " V" VERSION " (" __DATE2__ ")",

		MUIA_Application_Menustrip, MenustripObject,

			MUIA_Family_Child, MenuObjectT("Project"),
				MUIA_Family_Child, mi_Open = ReqMenuitemObject,
					MUIA_ReqMenuitem_ConfirmText,		"The file has changed,\nOk to open a new one?",
					MUIA_ReqMenuitem_Mode,				MODE_OLDFILE,
					MUIA_Menuitem_Title,					"Open...",
					MUIA_Menuitem_Shortcut,				"O",
					ASLFR_SleepWindow,					TRUE,
					End,
				MUIA_Family_Child, MenuitemObject, MUIA_Menuitem_Title, -1, End,
				MUIA_Family_Child, mi_Save = MenuitemObject,
					MUIA_Menuitem_Title,					"Save",
					MUIA_Menuitem_Shortcut,				"S",
					End,
				MUIA_Family_Child, mi_SaveAs = ReqMenuitemObject,
					MUIA_ReqMenuitem_Confirm,			TRUE,
					MUIA_ReqMenuitem_Mode,				MODE_NEWFILE,
					MUIA_Menuitem_Title,					"Save As...",
					MUIA_Menuitem_Shortcut,				"A",
					ASLFR_DoSaveMode,						TRUE,
					ASLFR_SleepWindow,					TRUE,
					End,
				MUIA_Family_Child, MenuitemObject, MUIA_Menuitem_Title, -1, End,
				MUIA_Family_Child, mi_Quit = MenuitemObject,
					MUIA_Menuitem_Title,					"Quit",
					MUIA_Menuitem_Shortcut,				"Q",
					End,
				End, /* Menu */
			End, /* MenuStrip */

		SubWindow, win = WindowObject,

			MUIA_Window_Activate,		TRUE,
			MUIA_Window_CloseGadget,	TRUE,
			MUIA_Window_DepthGadget,	TRUE,
			MUIA_Window_DragBar,			TRUE,
			MUIA_Window_SizeGadget,		TRUE,
			MUIA_Window_Title,			PROGRAM " V" VERSION,
			MUIA_Window_ID,				'MAIN',

			WindowContents, VGroup,

				Child, text = TextObject, TextFrame,
					MUIA_Background, MUII_TextBack,
					End,

				Child, HGroup,
					MUIA_Group_Spacing, 0,
					Child, texteditor = TextEditorObject, End,
					Child, slider = ScrollbarObject, End,
					End,

				End, /* VGroup */

			End, /* Window */

		End; /* Application */


	if(app)
	{
		/* Setup notifications */
		DoMethod(win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);
		DoMethod(mi_Open, MUIM_Notify, MUIA_ReqMenuitem_Lock, MUIV_EveryTime, texteditor, 3, MUIM_CallHook, &LoadHook, MUIV_TriggerValue);
		DoMethod(mi_Open, MUIM_Notify, MUIA_ReqMenuitem_Path, MUIV_EveryTime, mi_SaveAs, 3, MUIM_NoNotifySet, MUIA_ReqMenuitem_Path, MUIV_TriggerValue);
		DoMethod(mi_Open, MUIM_Notify, MUIA_ReqMenuitem_File, MUIV_EveryTime, text, 4, MUIM_SetAsString, MUIA_Text_Contents, "%s", MUIV_TriggerValue);
		DoMethod(mi_Save, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime, mi_SaveAs, 2, MUIM_ReqMenuitem_Trigger, MUIF_ReqMenuitem_Trigger_Quiet);
		DoMethod(mi_SaveAs, MUIM_Notify, MUIA_ReqMenuitem_Lock, MUIV_EveryTime, texteditor, 3, MUIM_CallHook, &SaveHook, MUIV_TriggerValue);
		DoMethod(mi_SaveAs, MUIM_Notify, MUIA_ReqMenuitem_Path, MUIV_EveryTime, mi_Open, 3, MUIM_NoNotifySet, MUIA_ReqMenuitem_Path, MUIV_TriggerValue);
		DoMethod(mi_Quit, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);
		DoMethod(texteditor, MUIM_Notify, MUIA_TextEditor_HasChanged, MUIV_EveryTime, mi_Open, 3, MUIM_Set, MUIA_ReqMenuitem_Confirm, MUIV_TriggerValue);

		/* Set various attributes */
		set(texteditor, MUIA_TextEditor_Slider, slider);
		SetAttrs(win,
			MUIA_Window_ActiveObject, texteditor,
			MUIA_Window_Open, TRUE,
			TAG_DONE);

		BOOL retry;
		do {

			retry = FALSE;
			ULONG sigs = 0;
			while((DoMethod(app, MUIM_Application_NewInput, &sigs) != MUIV_Application_ReturnID_Quit))
			{
				if(sigs)
				{
					if((sigs = Wait(sigs | SIGBREAKF_CTRL_C)) & SIGBREAKF_CTRL_C)
						break;
				}
			}

			ULONG changed;
			if(get(texteditor, MUIA_TextEditor_HasChanged, &changed), changed)
			{
				enum { Cancel = 0, Proceed, Save };
				switch(MUI_Request(app, win, 0L, "Warning", "*_Proceed|_Save|_Cancel", "\33cText has been modified. Save it?"))
				{
					case Cancel:
						retry = TRUE;
					break;

					case Save:
						retry = !DoMethod(mi_SaveAs, MUIM_ReqMenuitem_Trigger, MUIF_ReqMenuitem_Trigger_Quiet);
					break;
				}
			}

		} while(retry);

		MUI_DisposeObject(app);
	}
}

void wbmain (struct WBStartup *) { main(); };
